/*	/share/epicsH  $Id: seq.h,v 1.11 1993/08/11 12:50:05 cha Exp $
 *
 *	DESCRIPTION: Definitions for the run-time sequencer.
 *
 *      Author:         Andy Kozubal
 *      Date:           
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991,2,3, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * 07mar91,ajk	Changed SSCB semaphore id names.
 * 05jul91,ajk	Added function prototypes.
 * 16dec91,ajk	Setting up for VxWorks version 5.
 * 27apr92,ajk	Changed to new event flag mode (SSCB & PROG changed).
 * 27apr92,ajk	Removed getSemId from CHAN.
 * 28apr92,ajk	Implemented new event flag mode.
 * 17feb93,ajk	Fixed some functions prototypes.
 * 10jun93,ajk	Removed VxWorks V4/V5 conditional compile.
 * 20jul93,ajk	Removed non-ANSI function definitions.
 * 19apr93,bkc	Modified to run on Unix as well as on vxWorks.
 */


#ifndef INCseqh
#define INCseqh

#ifdef vxWorks
#ifndef	TRUE
#include	<ctype.h>
#endif
#ifndef	VOID
#include	"cadef.h"
#include	"db_access.h"
#include	"alarm.h"
#include	"vxWorks.h"
#include	"ioLib.h"
#include	"semLib.h"
#include	"taskLib.h"
#endif
/* Macros to handle set & clear event bits */
#define NBITS           32		/* # bits in bit mask word */
#define	NWRDS		8		/* # words in bit fields */
typedef long            bitMask;

#define bitSet(word, bitnum)	(word[bitnum/NBITS] |= (1<<(bitnum%NBITS)))
#define bitClear(word, bitnum)	(word[bitnum/NBITS] &= ~(1<<(bitnum%NBITS)))
#define bitTest(word, bitnum)	((word[bitnum/NBITS] & (1<<(bitnum%NBITS))) != 0)

/* Bit encoding for run-time options */
#define	OPT_DEBUG	(1<<0)
#define	OPT_ASYNC	(1<<1)
#define	OPT_CONN	(1<<2)
#define	OPT_REENT	(1<<3)
#define	OPT_NEWEF	(1<<4)
/* Structure to hold information about database channels */
struct	db_channel
{
	short		index;		/* index in array of db channels */
	char		*db_uname;	/* unexpanded channel name */
	char		*db_name;	/* channel name after macro expansion */
	char		*var;		/* ptr to variable (or an offset) */
	chid		chid;		/* ptr to channel id (from ca_search()) */
	BOOL		connected;	/* TRUE only if channel is connected */
	BOOL		get_complete;	/* TRUE if previous pvGet completed */
	short		status;		/* last db access status code */
	short		severity;	/* last db access severity code */
	short		size;		/* size (in bytes) of single variable element */
	short		get_type;	/* db get type (e.g. DBR_STS_INT) */
	short		put_type;	/* db put type (e.g. DBR_INT) */
	int		count;		/* element count for db requests */
	BOOL		mon_flag;	/* TRUE if channel is to be monitored */
	BOOL		monitored;	/* TRUE if channel IS monitored */
	float		delta;		/* monitor dead-band */
	float		timeout;	/* monitor timeout */
	evid		evid;		/* event id (supplied by CA) */
	SEM_ID		getSemId;	/* semaphore id for async get */
	struct state_program *sprog;	/* state program that owns this structure */
};
typedef	struct db_channel CHAN;

/* Structure to hold information about a state */
struct	state_info_block
{
	FUNCPTR		action_func;	/* ptr to action routine for this state */
	FUNCPTR		event_func;	/* ptr to event routine for this state */
	FUNCPTR		delay_func;	/* ptr to delay routine for this state */
	bitMask		eventMask[NWRDS];/* event mask for this state */
	char		*name;		/* state name (for debugging) */
};
typedef	struct	state_info_block STATE;

#define	MAX_NDELAY	20		/* max # delays allowed in each SS */
/* Structure to hold information about a State Set */
struct	state_set_control_block
{
	int		task_id;	/* task id */
	int		task_priority;	/* task priority */
	SEM_ID		syncSemId;	/* semaphore for event sync */
	SEM_ID		getSemId;	/* semaphore for synchronous pvGet() */
	int		num_states;	/* number of states */
	STATE		*states;	/* ptr to array of state blocks */
	short		current_state;	/* current state index */
	short		next_state;	/* next state index */
	short		prev_state;	/* previous state index */
	short		error_state;	/* error state index (-1 if none defined) */
	short		trans_num;	/* highest priority trans. # that triggered */
	BOOL		action_complete;/* TRUE if action complete
					   FALSE if waiting for an event */
	bitMask		*pMask;		/* current event mask */
	int		ndelay;		/* number of delays activated */
	ULONG		timeout[MAX_NDELAY]; /* time-out for delay() */
	ULONG		time;		/* time that a state was entered */
	char		*name;		/* state set name (for debugging) */
};
typedef	struct	state_set_control_block SSCB;

/* Macro table */
typedef	struct	macro {
	char	*name;
	char	*value;
} MACRO;

/* All information about a state program.
	The address of this structure is passed to the run-time sequencer:
 */
struct	state_program
{
	int		magic;		/* magic number so we don't s... on system */
	int		task_id;	/* task id (main task) */
	char		*dyn_ptr;	/* ptr to allocated area at run time */
	BOOL		task_is_deleted;/* TRUE if main task has been deleted */
	int		task_priority;	/* task priority */
	SEM_ID		caSemId;	/* semiphore for locking CA events */
	CHAN		*channels;	/* table of channels */
	int		nchan;		/* number of db channels */
	int		conn_count;	/* number of channels connected */
	SSCB		*sscb;		/* array of state set control blocks */
	int		nss;		/* number of state sets */
	STATE		*states;	/* ptr to dyn array of state blocks */
	int		nstates;	/* number of states */
	char		*user_area;	/* ptr to user variable area */
	int		user_size;	/* # bytes in user variable area */
	MACRO		*mac_ptr;	/* ptr to macro table */
	char		*scr_ptr;	/* ptr to start of scratch area */
	int		scr_nleft;	/* # bytes left in scratch area */
	char		*name;		/* program name (for debugging) */
	char		*params;	/* program paramters */
	bitMask		events[NWRDS];	/* global event bits */
	bitMask		options;	/* options (bit-encoded) */
	FUNCPTR		exit_func;	/* exit function */
	SEM_ID		logSemId;	/* logfile locking semaphore */
	int		logFd;		/* logfile file descr. */
};
typedef	struct	state_program SPROG;

#define	MAX_MACROS	20

#define	MAGIC 920505			/* magic number for SPROG */

/* Macros for special function calls */
#define	seq_params		sprog, ss_ptr

#define	pvPut(db_index)		seq_pvPut(seq_params, &sprog->channels[db_index])

#define	pvGet(db_index)		seq_pvGet(seq_params, &sprog->channels[db_index])

#define	pvMonitor(db_index)	seq_pvMonitor(seq_params, &sprog->channels[db_index])

#define	pvStopMonitor(db_index)	seq_pvStopMonitor(seq_params, &sprog->channels[db_index])

#define	pvStatus(db_index)	(sprog->channels[db_index].status)

#define	pvSeverity(db_index)	(sprog->channels[db_index].severity)

#define	pvConnected(db_index)	(sprog->channels[db_index].connected)

#define	pvGetComplete(db_index)	(sprog->channels[db_index].get_complete)

#define	pvChannelCount()	(sprog->nchan)

#define	pvConnectCount()	(sprog->conn_count)

#define	pvFlush()		seq_pvFlush(seq_params)

#define	pvCount(db_index)	(sprog->channels[db_index].count)

#define	pvIndex(db_index)	(db_index)

#define	efSet(ef_name)		seq_efSet(seq_params, ef_name)

#define	efTest(ef_name)		seq_efTest(seq_params, ef_name)

#define	efClear(ef_name)	(VOID)seq_efClear(seq_params, ef_name)

#define	efTestAndClear(ef_name)	seq_efClear(seq_params, ef_name)

#define	test_delay(delay_id)	seq_test_delay(seq_params, delay_id)

#define	start_delay(delay_id, delay)\
	seq_start_delay(seq_params, delay_id, (float)delay)

#define	macValueGet(mac_name)\
	seqMacValGet(mac_name, strlen(mac_name), sprog->mac_ptr)

#define	optGet(opt)		seq_optGet(sprog, opt)

/* Task parameters */
#define SPAWN_STACK_SIZE	10000
#define SPAWN_OPTIONS		VX_DEALLOC_STACK | VX_FP_TASK | VX_STDIO
#define SPAWN_PRIORITY		100

/* Function declarations */
int	seq_connect(SPROG *);
VOID	seq_event_handler(struct event_handler_args);
VOID	seq_conn_handler(struct connection_handler_args);
VOID	seq_efSet(SPROG *, int, int);
int	seq_efTest(SPROG *, SSCB *, int);
int	seq_efClear(SPROG *, SSCB *, int);
int	seq_pvGet(SPROG *, SSCB *, CHAN *);
VOID	seq_callback_handler(struct event_handler_args);
VOID	seq_pvFlush();
int	seq_pvPut(SPROG *, SSCB *, CHAN *);
int	seq_pvMonitor(SPROG *, SSCB *, CHAN *);
int	seq_pvStopMonitor(SPROG *, SSCB *, CHAN *);
int	seq(SPROG *, char *, int);
char	*seqAlloc(SPROG *, int);
STATUS	seqLog();
int	sequencer(SPROG *, int, char *);
VOID	ss_entry(SPROG *, SSCB *);
VOID	seq_start_delay(SPROG *, SSCB *,int, double);
BOOL	seq_test_delay(SPROG *, SSCB *,int);
int	sprog_delete(int);
char	*seqMacValGet(char *, int, MACRO *);
VOID	seq_log();
#else /* !vxWorks */

#include	<ctype.h>

#include	"cadef.h"
#include	"db_access.h"
#include	"alarm.h"


#include "seq_taskLib.h"
typedef struct sema_id {
        int flag;
        int count;
        pthread_mutex_t mtx;
        pthread_cond_t con;
} SEM_TH;                      /* semaphore defined for pthread */

typedef SEM_TH *SEM_ID;


/* Macros to handle set & clear event bits */
#define NBITS           32		/* # bits in bit mask word */
#define	NWRDS		8		/* # words in bit fields */
typedef long            bitMask;

#define bitSet(word, bitnum)	(word[bitnum/NBITS] |= (1<<(bitnum%NBITS)))
#define bitClear(word, bitnum)	(word[bitnum/NBITS] &= ~(1<<(bitnum%NBITS)))
#define bitTest(word, bitnum)	((word[bitnum/NBITS] & (1<<(bitnum%NBITS))) != 0)

/* copy from the vw/h/vxWorks.h */

#define LOCAL static
#if     !defined(NULL)
#define NULL            0
#endif

#if     !defined(EOF) || (EOF!=(-1))
#define EOF             (-1)
#endif

#if     !defined(FALSE) || (FALSE!=0)
#define FALSE           0
#endif

#if     !defined(TRUE) || (TRUE!=1)
#define TRUE            1
#endif


#define NONE            (-1)    /* for times when NULL won't do */
#define EOS             '\0'    /* C string terminator */
/* return status values */

#define OK              0
#define ERROR           (-1)

/* timeout defines */

#define NO_WAIT         0

/* low-level I/O input, output, error fd's */

#define STD_IN  0
#define STD_OUT 1
#define STD_ERR 2


/* copy from the vw/h/ioLib.h */

/* ioctl function codes */
#define FIOGETNAME      18              /* return file name in arg */
#define FIOSYNC         21              /* sync to disk */

/* copy from the vw/h/vxTypes.h */

typedef int             BOOL;
typedef void            VOID;
typedef unsigned long   ULONG;
/*typedef int             STATUS; */
typedef int             (*FUNCPTR) ();     /* ptr to function returning int */

/* copy from the vw/h/taskLib.h */

/* task option bits */

#define VX_SUPERVISOR_MODE      0x01    /* OBSOLETE: tasks always in sup mode */
#define VX_UNBREAKABLE          0x02    /* INTERNAL: breakpoints ignored */
#define VX_DEALLOC_STACK        0x04    /* INTERNAL: deallocate stack */
#define VX_FP_TASK              0x08    /* 1 = f-point coprocessor support */
#define VX_STDIO                0x10    /* 1 = stdioLib support */
#define VX_ADA_DEBUG            0x20    /* 1 = VADS debugger support */
#define VX_FORTRAN              0x40    /* 1 = NKR FORTRAN support */
#define VX_PRIVATE_ENV          0x80    /* 1 = private environment variables */


/* Bit encoding for run-time options */

#define	OPT_DEBUG	(1<<0)
#define	OPT_ASYNC	(1<<1)
#define	OPT_CONN	(1<<2)
#define	OPT_REENT	(1<<3)
#define	OPT_NEWEF	(1<<4)

/* Structure to hold information about database channels */
struct	db_channel
{
	short		index;		/* index in array of db channels */
	char		*db_uname;	/* unexpanded channel name */
	char		*db_name;	/* channel name after macro expansion */
	char		*var;		/* ptr to variable (or an offset) */
	chid		chid;		/* ptr to channel id (from ca_search()) */
	BOOL		connected;	/* TRUE only if channel is connected */
	BOOL		get_complete;	/* TRUE if previous pvGet completed */
	short		status;		/* last db access status code */
	short		severity;	/* last db access severity code */
	short		size;		/* size (in bytes) of single variable element */
	short		get_type;	/* db get type (e.g. DBR_STS_INT) */
	short		put_type;	/* db put type (e.g. DBR_INT) */
	int		count;		/* element count for db requests */
	BOOL		mon_flag;	/* TRUE if channel is to be monitored */
	BOOL		monitored;	/* TRUE if channel IS monitored */
	float		delta;		/* monitor dead-band */
	float		timeout;	/* monitor timeout */
	evid		evid;		/* event id (supplied by CA) */
	SEM_ID		getSemId;	/* semaphore id for async get */
	struct state_program *sprog;	/* state program that owns this structure */
};
typedef	struct db_channel CHAN;

/* Structure to hold information about a state */
struct	state_info_block
{
	FUNCPTR		action_func;	/* ptr to action routine for this state */
	FUNCPTR		event_func;	/* ptr to event routine for this state */
	FUNCPTR		delay_func;	/* ptr to delay routine for this state */
	bitMask		eventMask[NWRDS];/* event mask for this state */
	char		*name;		/* state name (for debugging) */
};
typedef	struct	state_info_block STATE;

#define	MAX_NDELAY	20		/* max # delays allowed in each SS */
/* Structure to hold information about a State Set */
struct	state_set_control_block
{
	int		task_id;	/* task id */
	int		task_priority;	/* task priority */
	SEM_ID		syncSemId;	/* semaphore for event sync */
	SEM_ID		getSemId;	/* semaphore for synchronous pvGet() */
	int		num_states;	/* number of states */
	STATE		*states;	/* ptr to array of state blocks */
	short		current_state;	/* current state index */
	short		next_state;	/* next state index */
	short		prev_state;	/* previous state index */
	short		error_state;	/* error state index (-1 if none defined) */
	short		trans_num;	/* highest priority trans. # that triggered */
	BOOL		action_complete;/* TRUE if action complete
					   FALSE if waiting for an event */
	bitMask		*pMask;		/* current event mask */
	int		ndelay;		/* number of delays activated */
	ULONG		timeout[MAX_NDELAY]; /* time-out for delay() */
	ULONG		time;		/* time that a state was entered */
	char		*name;		/* state set name (for debugging) */
};
typedef	struct	state_set_control_block SSCB;

/* Macro table */
typedef	struct	macro {
	char	*name;
	char	*value;
} MACRO;

/* All information about a state program.
	The address of this structure is passed to the run-time sequencer:
 */
struct	state_program
{
	int		magic;		/* magic number so we don't s... on system */
	int		task_id;	/* task id (main task) */
	char		*dyn_ptr;	/* ptr to allocated area at run time */
	BOOL		task_is_deleted;/* TRUE if main task has been deleted */
	int		task_priority;	/* task priority */
	SEM_ID		caSemId;	/* semiphore for locking CA events */
	CHAN		*channels;	/* table of channels */
	int		nchan;		/* number of db channels */
	int		conn_count;	/* number of channels connected */
	SSCB		*sscb;		/* array of state set control blocks */
	int		nss;		/* number of state sets */
	STATE		*states;	/* ptr to dyn array of state blocks */
	int		nstates;	/* number of states */
	char		*user_area;	/* ptr to user variable area */
	int		user_size;	/* # bytes in user variable area */
	MACRO		*mac_ptr;	/* ptr to macro table */
	char		*scr_ptr;	/* ptr to start of scratch area */
	int		scr_nleft;	/* # bytes left in scratch area */
	char		*name;		/* program name (for debugging) */
	char		*params;	/* program paramters */
	bitMask		events[NWRDS];	/* global event bits */
	bitMask		options;	/* options (bit-encoded) */
	FUNCPTR		exit_func;	/* exit function */
	SEM_ID		logSemId;	/* logfile locking semaphore */
	int		logFd;		/* logfile file descr. */
};
typedef	struct	state_program SPROG;

#define	MAX_MACROS	20

char	*seqAlloc(), *seqMacValGet();

#define	MAGIC 920505			/* magic number for SPROG */

/* Macros for special function calls */
#define	seq_params		sprog, ss_ptr

#define	pvPut(db_index)		seq_pvPut(seq_params, &sprog->channels[db_index])

#define	pvGet(db_index)		seq_pvGet(seq_params, &sprog->channels[db_index])

#define	pvMonitor(db_index)	seq_pvMonitor(seq_params, &sprog->channels[db_index])

#define	pvStopMonitor(db_index)	seq_pvStopMonitor(seq_params, &sprog->channels[db_index])

#define	pvStatus(db_index)	(sprog->channels[db_index].status)

#define	pvSeverity(db_index)	(sprog->channels[db_index].severity)

#define	pvConnected(db_index)	(sprog->channels[db_index].connected)

#define	pvGetComplete(db_index)	(sprog->channels[db_index].get_complete)

#define	pvChannelCount()	(sprog->nchan)

#define	pvConnectCount()	(sprog->conn_count)

#define	pvFlush()		seq_pvFlush(seq_params)

#define	pvCount(db_index)	(sprog->channels[db_index].count)

#define	pvIndex(db_index)	(db_index)

#define	efSet(ef_name)		seq_efSet(seq_params, ef_name)

#define	efTest(ef_name)		seq_efTest(seq_params, ef_name)

#define	efClear(ef_name)	(VOID)seq_efClear(seq_params, ef_name)

#define	efTestAndClear(ef_name)	seq_efClear(seq_params, ef_name)

#define	test_delay(delay_id)	seq_test_delay(seq_params, delay_id)

#define	start_delay(delay_id, delay)\
	seq_start_delay(seq_params, delay_id, (float)delay)

#define	macValueGet(mac_name)\
	seqMacValGet(mac_name, strlen(mac_name), sprog->mac_ptr)

#define	optGet(opt)		seq_optGet(sprog, opt)

/* Task parameters */
#define SPAWN_STACK_SIZE	10000
#define SPAWN_OPTIONS		VX_DEALLOC_STACK | VX_FP_TASK | VX_STDIO
#define SPAWN_PRIORITY		100


#define	semTake(semId, timeout)	seq_semTake(semId, timeout)

#define	SEM_INVERSION_SAFE	0
#define	SEM_DELETE_SAFE		0
#define	SEM_FULL		2	/* constant defined for semaphore*/
#define	SEM_EMPTY		1	/* constant defined for semaphore*/
#define	SEM_Q_PRIORITY		0
#define	SEM_Q_FIFO		0
#define	NO_WAIT			0
#define	WAIT_FOREVER		(-1)	/* wait for ever */

/*#define	ANSI*/

/* Function declarations */
#ifdef	ANSI
/* ANSI function prototypes */
int	seq_connect(SPROG *);
VOID	seq_event_handler(struct event_handler_args);
VOID	seq_conn_handler(struct connection_handler_args);
VOID	seq_efSet(SPROG *, int, int);
int	seq_efTest(SPROG *, SSCB *, int);
int	seq_efClear(SPROG *, SSCB *, int);
int	seq_pvGet(SPROG *, SSCB *, CHAN *);
VOID	seq_callback_handler(struct event_handler_args);
VOID	seq_pvFlush();
int	seq_pvPut(SPROG *, SSCB *, CHAN *);
int	seq_pvMonitor(SPROG *, SSCB *, CHAN *);
int	seq_pvStopMonitor(SPROG *, SSCB *, CHAN *);
int	seq(SPROG *, char *, int);
char	*seqAlloc(SPROG *, int);
STATUS	seqLog();
int	sequencer(SPROG *, int, char *);
VOID	ss_entry(SPROG *, SSCB *);
VOID	seq_start_delay(SPROG *, SSCB *,int, double);
BOOL	seq_test_delay(SPROG *, SSCB *,int);
int	sprog_delete(TCBX *);
#ifndef	V5_vxWorks
#ifndef FSUTHREAD
SEM_ID  semMCreate(int);
SEM_ID  semBCreate(int);
#endif	FSUTHREAD
#endif	V5_vxWorks

#else
/* Archaic (i.e. Sun) function prototypes */
int	seq_connect();
VOID	seq_event_handler();
VOID	seq_conn_handler();
VOID	seq_efSet();
int	seq_efTest();
int	seq_efClear();
int	seq_pvGet();
VOID	seq_callback_handler();
VOID	seq_pvFlush();
int	seq_pvPut();
int	seq_pvMonitor();
int	seq_pvStopMonitor();
int	seq();
char	*seqAlloc();
STATUS	seqLog();
int	sequencer();
VOID	ss_entry();
VOID	seq_start_delay();
BOOL	seq_test_delay();
int	sprog_delete();
#ifndef	V5_vxWorks
SEM_ID	semMCreate();
SEM_ID	semBCreate();
#endif	V5_vxWorks
#endif	ANSI

#endif /* vxWorks */

#endif /* INCseqh */
